"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const core_1 = require("@oclif/core");
const node_fs_1 = tslib_1.__importDefault(require("node:fs"));
const node_path_1 = tslib_1.__importDefault(require("node:path"));
const safe_stable_stringify_1 = tslib_1.__importDefault(require("safe-stable-stringify"));
const generator_js_1 = require("../factory/generator.js");
const BaseError_js_1 = require("../src/Error/BaseError.js");
class Generate extends core_1.Command {
    static description = "Generate JSON schema from your Typescript sources";
    static examples = [
        {
            command: "<%= config.bin %> <%= command.id %> src/types.ts -p tsconfig.json",
            description: "Analyzes src/types.ts using tsconfig.json and pipes to stdout",
        },
        {
            command: "<%= config.bin %> <%= command.id %> src/types.ts -o schema.json",
            description: "Analyzes src/types.ts and writes the schema to schema.json",
        },
    ];
    static args = {
        path: core_1.Args.file({
            description: "Source root filepath",
            required: true,
        }),
    };
    static flags = {
        type: core_1.Flags.string({
            char: "t",
            description: "Type name",
        }),
        "top-ref": core_1.Flags.boolean({
            description: "Create a top-level $ref definition",
            default: true,
            allowNo: true,
        }),
        id: core_1.Flags.string({
            char: "i",
            description: "$id for generated schema",
            aliases: ["schema-id"],
        }),
        tsconfig: core_1.Flags.file({
            char: "p",
            aliases: ["project"],
            description: "Your tsconfig.json to load entry files and compilation settings",
            default: "tsconfig.json",
        }),
        expose: core_1.Flags.string({
            char: "e",
            description: "Type exposing",
            options: ["all", "none", "export"],
            default: "export",
        }),
        jsdoc: core_1.Flags.string({
            char: "j",
            aliases: ["jsDoc"],
            description: "Read JsDoc annotations",
            options: ["none", "basic", "extended"],
            default: "extended",
        }),
        "markdown-description": core_1.Flags.boolean({
            description: "Generate `markdownDescription` in addition to `description`. Implies --jsdoc=extended",
            default: false,
        }),
        "sort-props": core_1.Flags.boolean({
            description: "Makes the schema stable by sorting properties",
            default: true,
            allowNo: true,
        }),
        "strict-tuples": core_1.Flags.boolean({
            description: "Do not allow additional items on tuples",
            default: false,
        }),
        "type-check": core_1.Flags.boolean({
            description: "Type checks to improve performance",
            default: true,
            allowNo: true,
        }),
        "ref-encode": core_1.Flags.boolean({
            description: "Encode references",
            default: true,
            allowNo: true,
        }),
        "additional-properties": core_1.Flags.boolean({
            description: "Allow additional properties for objects with no index signature",
            default: false,
        }),
        functions: core_1.Flags.string({
            description: "How to handle functions. `fail` will throw an error. `comment` will add a comment. `hide` will treat the function like a NeverType or HiddenType.",
            options: ["fail", "comment", "hide"],
            default: "comment",
        }),
        minify: core_1.Flags.boolean({
            description: "Minify generated schema",
            default: false,
            exclusive: ["color"],
        }),
        color: core_1.Flags.boolean({
            description: "Pretty print the json with colors. Only works when outputting to stdout. Defaults to true when TTY",
            default: async (c) => !c.flags.minify && (!!process.env.TTY || process.stdout.isTTY),
            allowNo: true,
            exclusive: ["minify"],
        }),
        out: core_1.Flags.file({
            char: "o",
            description: "Set the output file (default: stdout)",
        }),
        "extra-tags": core_1.Flags.string({
            description: "Provide additional validation keywords to include",
            multiple: true,
            aliases: ["validation-keywords"],
            default: [],
        }),
        "discriminator-type": core_1.Flags.string({
            description: "Type of discriminator to use",
            options: ["json-schema", "open-api"],
            default: "json-schema",
        }),
    };
    async run() {
        const { flags, args } = await this.parse(Generate);
        if (flags["markdown-description"]) {
            flags.jsdoc = "extended";
        }
        if (flags["type-check"] && !flags.tsconfig) {
            if (flags.out) {
                this.warn("Cannot type check without a tsconfig file. Skipping type check.");
            }
            flags["type-check"] = false;
        }
        const config = {
            minify: flags.minify,
            path: args.path,
            type: flags.type,
            expose: flags.expose,
            jsDoc: flags.jsdoc,
            markdownDescription: flags.markdownDescription,
            sortProps: flags["sort-props"],
            strictTuples: flags["strict-tuples"],
            skipTypeCheck: !flags["type-check"],
            additionalProperties: flags["additional-properties"],
            discriminatorType: flags["discriminator-type"],
            encodeRefs: flags["ref-encode"],
            extraTags: flags["extra-tags"],
            functions: flags.functions,
            topRef: flags["top-ref"],
            schemaId: flags.id,
            tsconfig: flags.tsconfig,
        };
        try {
            const generator = (0, generator_js_1.createGenerator)(config);
            const { rootNodes, schema } = generator.createSchema(config.type);
            const stringify = config.sortProps ? safe_stable_stringify_1.default : JSON.stringify;
            const schemaString = (config.minify ? stringify(schema) : stringify(schema, null, 2));
            if (flags.out) {
                const outPath = node_path_1.default.dirname(flags.out);
                await node_fs_1.default.promises.mkdir(outPath, { recursive: true });
                await node_fs_1.default.promises.writeFile(flags.out, schemaString, { encoding: "utf-8" });
                this.log(`Schema of ${rootNodes.length} root nodes written to .${node_path_1.default.sep}${node_path_1.default.relative(process.cwd(), flags.out)}`);
                return;
            }
            if (flags.color) {
                this.logJson(schema);
            }
            else {
                this.log(schemaString);
            }
        }
        catch (error) {
            if (error instanceof BaseError_js_1.BaseError && error.diagnostic.file) {
                this.log(error.format());
                this.exit(1);
            }
            this.error(error);
        }
    }
}
exports.default = Generate;
//# sourceMappingURL=generate.js.map